@@ -3,7 +3,7 @@
 /* header file for S-Lang internal structures that users do not (should not)
    need.  Use slang.h for that purpose. */
 /*
-Copyright (C) 2004-2007 John E. Davis
+Copyright (C) 2004-2009 John E. Davis
 
 This file is part of the S-Lang Library.
 
@@ -158,9 +158,9 @@ typedef enum
    SLANG_BC_RETURN		= 0x64,
    SLANG_BC_BREAK		= 0x65,
    SLANG_BC_CONTINUE		= 0x66,
-   SLANG_BC_IF_BLOCK		= 0x67,   /* optimized code */
-   SLANG_BC_UNUSED_0x68		= 0x68,
-   SLANG_BC_UNUSED_0x69		= 0x69,
+   SLANG_BC_UNUSED_0x67		= 0x67,
+   SLANG_BC_CONTINUE_N		= 0x68,
+   SLANG_BC_BREAK_N		= 0x69,
    SLANG_BC_X_ERROR		= 0x6A,
    SLANG_BC_X_USER0		= 0x6B,
    SLANG_BC_X_USER1		= 0x6C,
@@ -222,32 +222,30 @@ typedef enum
    SLANG_BC_MANY_LVARIABLE_DIR	= 0x9D,
    SLANG_BC_LVARIABLE_AGET1	= 0x9E,
    SLANG_BC_LITERAL_AGET1	= 0x9F,
-   SLANG_BC_LVARIABLE_APUT1	= 0xA0,
-   SLANG_BC_LITERAL_APUT1	= 0xA1,
-   SLANG_BC_UNUSED_0xA2		= 0xA2,
-   SLANG_BC_UNUSED_0xA3		= 0xA3,
-   SLANG_BC_UNUSED_0xA4		= 0xA4,
-   SLANG_BC_UNUSED_0xA5		= 0xA5,
-   SLANG_BC_UNUSED_0xA6		= 0xA6,
-   SLANG_BC_UNUSED_0xA7		= 0xA7,
-   SLANG_BC_UNUSED_0xA8		= 0xA8,
-   SLANG_BC_UNUSED_0xA9		= 0xA9,
-   SLANG_BC_UNUSED_0xAA		= 0xAA,
-   SLANG_BC_UNUSED_0xAB		= 0xAB,
-   SLANG_BC_UNUSED_0xAC		= 0xAC,
-   SLANG_BC_UNUSED_0xAD		= 0xAD,
-   SLANG_BC_UNUSED_0xAE		= 0xAE,
-   SLANG_BC_UNUSED_0xAF		= 0xAF,
-   /* The following do not actually occur in inner_interp.  They used 
-    * to signify the bytecode has been combined with another.
-    */
-   SLANG_BC_LVARIABLE_COMBINED	= 0xB0,
-   SLANG_BC_GVARIABLE_COMBINED	= 0xB1,
-   SLANG_BC_LITERAL_COMBINED	= 0xB2,
-   SLANG_BC_CALL_DIRECT_COMB	= 0xB3,
-   SLANG_BC_COMBINED		= 0xB4,
-   SLANG_BC_UNUSED_0xB5		= 0xB5,
-   SLANG_BC_UNUSED_0xB6		= 0xB6,
+   SLANG_BC_LVAR_LVAR_APUT1	= 0xA0,
+   SLANG_BC_LVARIABLE_APUT1	= 0xA1,
+   SLANG_BC_LITERAL_APUT1	= 0xA2,
+   SLANG_BC_LLVARIABLE_BINARY2	= 0xA3,
+   SLANG_BC_SET_LOCLV_LIT_INT	= 0xA4,
+   SLANG_BC_SET_LOCLV_LIT_AGET1	= 0xA5,
+   SLANG_BC_SET_LOCLV_LVAR	= 0xA6,
+   SLANG_BC_SET_LOCLV_LASTBLOCK	= 0xA7,
+   SLANG_BC_LVAR_EARG_LVAR	= 0xA8,
+   SLANG_BC_LVAR_FIELD		= 0xA9,
+   SLANG_BC_BINARY_LASTBLOCK	= 0xAA,
+   SLANG_BC_EARG_LVARIABLE_INTRINSIC	= 0xAB,
+   SLANG_BC_LVAR_LITERAL_INT	= 0xAC,
+   SLANG_BC_BINARY_SET_LOCLVAL	= 0xAD,
+   SLANG_BC_LVAR_AGET_SET_LOCLVAL	= 0xAE,
+   SLANG_BC_LLVAR_BINARY_IF	= 0xAF,
+
+   SLANG_BC_IF_BLOCK		= 0xB0,
+   SLANG_BC_LVAR_SET_FIELD	= 0xB1,
+   SLANG_BC_PVAR_SET_GLOB_LVAL	= 0xB2,
+   SLANG_BC_LVAR_SET_GLOB_LVAL	= 0xB3,
+   SLANG_BC_LIT_AGET1_INT_BINARY= 0xB4,
+   SLANG_BC_BINARY2		= 0xB5,
+   SLANG_BC_LVAR_LIT_AGET1	= 0xB6,
    SLANG_BC_UNUSED_0xB7		= 0xB7,
    SLANG_BC_UNUSED_0xB8		= 0xB8,
    SLANG_BC_UNUSED_0xB9		= 0xB9,
@@ -257,12 +255,17 @@ typedef enum
    SLANG_BC_UNUSED_0xBD		= 0xBD,
    SLANG_BC_UNUSED_0xBE		= 0xBE,
    SLANG_BC_UNUSED_0xBF		= 0xBF,
-   SLANG_BC_UNUSED_0xC0		= 0xC0,
-   SLANG_BC_UNUSED_0xC1		= 0xC1,
-   SLANG_BC_UNUSED_0xC2		= 0xC2,
-   SLANG_BC_UNUSED_0xC3		= 0xC3,
-   SLANG_BC_UNUSED_0xC4		= 0xC4,
-   SLANG_BC_UNUSED_0xC5		= 0xC5,
+
+   /* The following do not actually occur in inner_interp.  They used 
+    * to signify the bytecode has been combined with another.
+    */
+   SLANG_BC_LVARIABLE_COMBINED	= 0xC0,
+   SLANG_BC_GVARIABLE_COMBINED	= 0xC1,
+   SLANG_BC_LITERAL_COMBINED	= 0xC2,
+   SLANG_BC_CALL_DIRECT_COMB	= 0xC3,
+   SLANG_BC_COMBINED		= 0xC4,
+   SLANG_BC_BLOCK_COMBINED	= 0xC5,
+#define SLANG_IS_BC_COMBINED(b) ((0xC0 <= (b)) && ((b) <= 0xC5))
    SLANG_BC_UNUSED_0xC6		= 0xC6,
    SLANG_BC_UNUSED_0xC7		= 0xC7,
    SLANG_BC_UNUSED_0xC8		= 0xC8,
@@ -428,14 +431,20 @@ typedef long _pSLc_off_t_Type;
 #if SIZEOF_INT == 2
 # define _pSLANG_INT16_TYPE	SLANG_INT_TYPE
 # define _pSLANG_UINT16_TYPE	SLANG_UINT_TYPE
+typedef int _pSLint16_Type;
+typedef unsigned int _pSLuint16_Type;
 #else
 # if SIZEOF_SHORT == 2
 #  define _pSLANG_INT16_TYPE	SLANG_SHORT_TYPE
 #  define _pSLANG_UINT16_TYPE	SLANG_USHORT_TYPE
+typedef short _pSLint16_Type;
+typedef unsigned short _pSLuint16_Type;
 # else
 #  if SIZEOF_LONG == 2
 #   define _pSLANG_INT16_TYPE	SLANG_LONG_TYPE
 #   define _pSLANG_UINT16_TYPE	SLANG_ULONG_TYPE
+typedef long _pSLInt16_Type;
+typedef unsigned long _pSLuint16_Type;
 #  else
 #   define _pSLANG_INT16_TYPE	0
 #   define _pSLANG_UINT16_TYPE	0
@@ -446,14 +455,20 @@ typedef long _pSLc_off_t_Type;
 #if SIZEOF_INT == 4
 # define _pSLANG_INT32_TYPE	SLANG_INT_TYPE
 # define _pSLANG_UINT32_TYPE	SLANG_UINT_TYPE
+typedef int _pSLint32_Type;
+typedef unsigned int _pSLuint32_Type;
 #else
 # if SIZEOF_SHORT == 4
 #  define _pSLANG_INT32_TYPE	SLANG_SHORT_TYPE
 #  define _pSLANG_UINT32_TYPE	SLANG_USHORT_TYPE
+typedef short _pSLInt32_Type;
+typedef unsigned short _pSLuint32_Type;
 # else
 #  if SIZEOF_LONG == 4
 #   define _pSLANG_INT32_TYPE	SLANG_LONG_TYPE
 #   define _pSLANG_UINT32_TYPE	SLANG_ULONG_TYPE
+typedef long _pSLInt32_Type;
+typedef unsigned long _pSLuint32_Type;
 #  else
 #   define _pSLANG_INT32_TYPE	0
 #   define _pSLANG_UINT32_TYPE	0
@@ -464,18 +479,26 @@ typedef long _pSLc_off_t_Type;
 #if SIZEOF_INT == 8
 # define _pSLANG_INT64_TYPE	SLANG_INT_TYPE
 # define _pSLANG_UINT64_TYPE	SLANG_UINT_TYPE
+typedef int _pSLint64_Type;
+typedef unsigned int _pSLuint64_Type;
 #else
 # if SIZEOF_SHORT == 8
 #  define _pSLANG_INT64_TYPE	SLANG_SHORT_TYPE
 #  define _pSLANG_UINT64_TYPE	SLANG_USHORT_TYPE
+typedef int _pSLInt64_Type;
+typedef unsigned int _pSLuint64_Type;
 # else
 #  if SIZEOF_LONG == 8
 #   define _pSLANG_INT64_TYPE	SLANG_LONG_TYPE
 #   define _pSLANG_UINT64_TYPE	SLANG_ULONG_TYPE
+typedef long _pSLInt64_Type;
+typedef unsigned long _pSLuint64_Type;
 #  else
 #   if SIZEOF_LONG_LONG == 8
 #    define _pSLANG_INT64_TYPE	SLANG_LLONG_TYPE
 #    define _pSLANG_UINT64_TYPE	SLANG_ULLONG_TYPE
+typedef long long _pSLInt64_Type;
+typedef unsigned long long _pSLuint64_Type;
 #   else
 #    define _pSLANG_INT64_TYPE	0
 #    define _pSLANG_UINT64_TYPE	0
@@ -515,7 +538,7 @@ _pSL_Object_Union_Type;
 
 typedef struct _pSLang_Object_Type
 {
-   SLtype data_type;	       /* SLANG_INT_TYPE, ... */
+   SLtype o_data_type;	       /* SLANG_INT_TYPE, ... */
    _pSL_Object_Union_Type v;
 }
 SLang_Object_Type;
@@ -531,7 +554,7 @@ extern int _pSLang_pop_object_of_type (SLtype, SLang_Object_Type *, int);
 
 typedef struct
 {
-   char *name;			       /* slstring */
+   SLFUTURE_CONST char *name;			       /* slstring */
    SLang_Object_Type obj;
 }
 _pSLstruct_Field_Type;
@@ -551,8 +574,10 @@ extern int _pSLstruct_define_struct (void);
 extern int _pSLstruct_define_struct2 (void);
 extern int _pSLstruct_define_typedef (void);
 
-extern SLang_Object_Type *_pSLstruct_get_field_value (SLang_Struct_Type *, char *);
-extern int _pSLstruct_push_field_ref (char *);
+extern SLang_Object_Type *_pSLstruct_get_field_value (SLang_Struct_Type *, SLCONST char *);
+extern int _pSLstruct_push_field_ref (SLFUTURE_CONST char *);
+extern int _pSLstruct_push_field (SLang_Struct_Type *s, SLFUTURE_CONST char *name, int do_free);
+extern int _pSLstruct_pop_field (SLang_Struct_Type *s, SLFUTURE_CONST char *name, int do_free);
 
 int _pSLang_get_qualifiers (SLang_Struct_Type **);
 
@@ -627,8 +652,8 @@ extern int _pSLstring_list_push (_pSLString_List_Type *, int);
 extern SLang_Array_Type *_pSLstrings_to_array (char **strs, unsigned int n);
 
 /* This function assumes that s is an slstring. */
-extern char *_pSLstring_dup_slstring (char *);
-extern int _pSLang_dup_and_push_slstring (char *);
+extern SLCONST char *_pSLstring_dup_slstring (SLCONST char *);
+extern int _pSLang_dup_and_push_slstring (SLCONST char *);
 
 
 extern int _pSLang_init_import (void);
@@ -636,15 +661,16 @@ extern int _pSLinit_exceptions (void);
 
 /* This function checks to see if the referenced object is initialized */
 extern int _pSLang_is_ref_initialized (SLang_Ref_Type *);
-extern int _pSLcheck_identifier_syntax (char *);
+extern int _pSLcheck_identifier_syntax (SLCONST char *);
 extern int _pSLang_uninitialize_ref (SLang_Ref_Type *);
 
 extern int _pSLpush_slang_obj (SLang_Object_Type *);
+extern int _pSLslang_copy_obj (SLang_Object_Type *obja, SLang_Object_Type *objb);
 
-extern char *_pSLexpand_escaped_char(char *, SLwchar_Type *, int *);
+extern char *_pSLexpand_escaped_char(char *, char *, SLwchar_Type *, int *);
 extern void _pSLexpand_escaped_string (char *, char *, char *);
 
-extern int _pSLpush_dollar_string (char *);
+extern int _pSLpush_dollar_string (SLFUTURE_CONST char *);
 
 /* returns a pointer to an SLstring string-- use SLang_free_slstring */
 extern char *_pSLstringize_object (SLang_Object_Type *);
@@ -654,38 +680,39 @@ extern SLang_Object_Type *_pSLang_get_run_stack_pointer (void);
 extern SLang_Object_Type *_pSLang_get_run_stack_base (void);
 extern int _pSLang_dump_stack (void);
 extern int _pSLang_peek_at_stack2 (SLtype *);
+extern int _pSLang_restart_arg_list (int nargs);
 
 struct _pSLang_NameSpace_Type
 {
    struct _pSLang_NameSpace_Type *next;
-   char *name;			       /* this is the load_type name */
-   char *namespace_name;	       /* this name is assigned by implements */
-   char *private_name;
+   SLFUTURE_CONST char *name;	       /* this is the load_type name */
+   SLFUTURE_CONST char *namespace_name;/* this name is assigned by implements */
+   SLFUTURE_CONST char *private_name;
    unsigned int table_size;
    SLang_Name_Type **table;
 };
-extern SLang_NameSpace_Type *_pSLns_new_namespace (char *, unsigned int);
-extern SLang_NameSpace_Type *_pSLns_allocate_namespace (char *, unsigned int);
+extern SLang_NameSpace_Type *_pSLns_new_namespace (SLFUTURE_CONST char *, unsigned int);
+extern SLang_NameSpace_Type *_pSLns_allocate_namespace (SLFUTURE_CONST char *, unsigned int);
 extern void _pSLns_deallocate_namespace (SLang_NameSpace_Type *);
-extern SLang_NameSpace_Type *_pSLns_find_namespace (char *);
-extern int _pSLns_set_namespace_name (SLang_NameSpace_Type *, char *);
-extern SLang_Array_Type *_pSLnspace_apropos (SLang_NameSpace_Type *, char *, unsigned int);
+extern SLang_NameSpace_Type *_pSLns_find_namespace (SLCONST char *);
+extern int _pSLns_set_namespace_name (SLang_NameSpace_Type *, SLFUTURE_CONST char *);
+extern SLang_Array_Type *_pSLnspace_apropos (SLang_NameSpace_Type *, SLFUTURE_CONST char *, unsigned int);
 extern void _pSLang_use_namespace_intrinsic (char *name);
-extern char *_pSLang_cur_namespace_intrinsic (void);
-extern SLang_Array_Type *_pSLang_apropos (char *, char *, unsigned int);
-extern void _pSLang_implements_intrinsic (char *);
+extern SLFUTURE_CONST char *_pSLang_cur_namespace_intrinsic (void);
+extern SLang_Array_Type *_pSLang_apropos (SLFUTURE_CONST char *, SLFUTURE_CONST char *, unsigned int);
+extern void _pSLang_implements_intrinsic (SLFUTURE_CONST char *);
 extern SLang_Array_Type *_pSLns_list_namespaces (void);
-extern SLang_Name_Type *_pSLns_locate_hashed_name (SLang_NameSpace_Type *, char *, unsigned long);
+extern SLang_Name_Type *_pSLns_locate_hashed_name (SLang_NameSpace_Type *, SLCONST char *, unsigned long);
 extern int _pSLns_add_hashed_name (SLang_NameSpace_Type *, SLang_Name_Type *, unsigned long);
 extern SLang_NameSpace_Type *_pSLns_find_object_namespace (SLang_Name_Type *nt);
-extern SLang_Name_Type *_pSLns_locate_name (SLang_NameSpace_Type *, char *);
-extern SLang_NameSpace_Type *_pSLns_get_private_namespace (char *name, char *nsname);
-extern SLang_NameSpace_Type *_pSLns_create_namespace2 (char *name, char *nsname);
+extern SLang_Name_Type *_pSLns_locate_name (SLang_NameSpace_Type *, SLCONST char *);
+extern SLang_NameSpace_Type *_pSLns_get_private_namespace (SLFUTURE_CONST char *name, SLFUTURE_CONST char *nsname);
+extern SLang_NameSpace_Type *_pSLns_create_namespace2 (SLFUTURE_CONST char *name, SLFUTURE_CONST char *nsname);
 
 extern int _pSLang_Trace;
-extern char *_pSLang_current_function_name (void);
+extern char SLCONST *_pSLang_current_function_name (void);
 
-extern int _pSLang_trace_fun(char *);
+extern int _pSLang_trace_fun(SLFUTURE_CONST char *);
 
 /* This is a bitmapped variable */
 /* extern int _pSLang_Compile_Line_Num_Info; */
@@ -693,26 +720,31 @@ extern int _pSLang_trace_fun(char *);
 extern int _pSLang_Compile_BOSEOS;
 extern int _pSLang_Compile_BOFEOF;
 extern int _pSLang_init_boseos (void);
-extern int _pSLcall_bos_handler (char *, int);
+extern int _pSLcall_bos_handler (SLFUTURE_CONST char *, int);
 extern int _pSLcall_eos_handler (void);
-extern int _pSLcall_bof_handler (char *, char *);
+extern int _pSLcall_bof_handler (SLFUTURE_CONST char *, SLFUTURE_CONST char *);
 extern int _pSLcall_eof_handler (void);
-extern int _pSLcall_debug_hook (char *file, int linenum);
+extern int _pSLcall_debug_hook (SLFUTURE_CONST char *file, int linenum);
 /* extern int _pSLcall_debug_hook (char *file, int linenum, char *funct); */
 #endif
 
-extern char *_pSLstring_dup_hashed_string (char *, unsigned long);
-extern unsigned long _pSLcompute_string_hash (char *);
-extern char *_pSLstring_make_hashed_string (char *, unsigned int, unsigned long *);
+extern char *_pSLstring_dup_hashed_string (SLCONST char *, unsigned long);
+extern unsigned long _pSLcompute_string_hash (SLCONST char *);
+extern char *_pSLstring_make_hashed_string (SLCONST char *, unsigned int, unsigned long *);
 extern void _pSLfree_hashed_string (char *, unsigned int, unsigned long);
-unsigned long _pSLstring_hash (unsigned char *, unsigned char *);
+unsigned long _pSLstring_hash (SLCONST unsigned char *, SLCONST unsigned char *);
 extern int _pSLinit_slcomplex (void);
 
 extern int _pSLang_init_slstrops (void);
 extern int _pSLstrops_do_sprintf_n (int);
 extern int _pSLang_sscanf (void);
-extern double _pSLang_atof (char *);
+extern double _pSLang_atof (SLFUTURE_CONST char *);
+
 extern int _pSLang_init_bstring (void);
+extern SLang_Foreach_Context_Type *_pSLbstring_foreach_open (SLtype type, unsigned int num);
+extern void _pSLbstring_foreach_close (SLtype type, SLang_Foreach_Context_Type *c);
+extern int _pSLbstring_foreach (SLtype type, SLang_Foreach_Context_Type *c);
+
 extern int _pSLang_init_sltime (void);
 extern void _pSLpack (void);
 extern void _pSLunpack (char *, SLang_BString_Type *);
@@ -729,7 +761,7 @@ extern int _pSLarith_typecast (SLtype, VOID_STAR, unsigned int,
 			      SLtype, VOID_STAR);
 
 extern int SLang_push(SLang_Object_Type *);
-extern int SLadd_global_variable (char *);
+extern int SLadd_global_variable (SLCONST char *);
 
 extern int _pSLdo_pop (void);
 extern unsigned int _pSLsys_getkey (void);
@@ -742,11 +774,11 @@ extern unsigned int _pSLpc_convert_scancode (unsigned int, unsigned int, int);
 #endif
 
 typedef struct _pSLterminfo_Type SLterminfo_Type;
-extern SLterminfo_Type *_pSLtt_tigetent (char *);
+extern SLterminfo_Type *_pSLtt_tigetent (SLCONST char *);
 extern void _pSLtt_tifreeent (SLterminfo_Type *);
-extern char *_pSLtt_tigetstr (SLterminfo_Type *, char *);
-extern int _pSLtt_tigetnum (SLterminfo_Type *, char *);
-extern int _pSLtt_tigetflag (SLterminfo_Type *, char *);
+extern char *_pSLtt_tigetstr (SLterminfo_Type *, SLCONST char *);
+extern int _pSLtt_tigetnum (SLterminfo_Type *, SLCONST char *);
+extern int _pSLtt_tigetflag (SLterminfo_Type *, SLCONST char *);
 
 #if SLTT_HAS_NON_BCE_SUPPORT
 extern int _pSLtt_get_bce_color_offset (void);
@@ -847,6 +879,12 @@ struct _pSLang_Class_Type
    int (*cl_datatype_deref)_PROTO((SLtype));
    SLang_Struct_Type *cl_struct_def;
    int (*cl_dereference) _PROTO((SLtype, VOID_STAR));
+
+   /* cl_a* functions deal with functions that have array addresses.  For 
+    * scalar and vector types, there is no difference.  However, for vector
+    * types such as Complex, which is stored as double[2], the address is
+    * the address of the first byte, and not the object itself.
+    */
    int (*cl_acopy) (SLtype, VOID_STAR, VOID_STAR);
    int (*cl_apop) _PROTO((SLtype, VOID_STAR));
    int (*cl_apush) _PROTO((SLtype, VOID_STAR));
@@ -871,8 +909,8 @@ struct _pSLang_Class_Type
    int (*cl_foreach) (SLtype, SLang_Foreach_Context_Type *);
 
    /* Structure access: get and put (assign to) fields */
-   int (*cl_sput) (SLtype, char *);
-   int (*cl_sget) (SLtype, char *);
+   int (*cl_sput) (SLtype, SLFUTURE_CONST char *);
+   int (*cl_sget) (SLtype, SLFUTURE_CONST char *);
 
    /* File I/O */
    int (*cl_fread) (SLtype, FILE *, VOID_STAR, unsigned int, unsigned int *);
@@ -884,10 +922,17 @@ struct _pSLang_Class_Type
    
    int (*cl_cmp)(SLtype, VOID_STAR, VOID_STAR, int *);   
    int (*cl_eqs)(SLtype, VOID_STAR, SLtype, VOID_STAR);
+   
+   /* required for any object that takes advantage of __tmp optimization */
+   void (*cl_inc_ref)(SLtype, VOID_STAR, int);
+
    SL_OOBinary_Type *cl_void_binary_this;
    SL_OOBinary_Type *cl_this_binary_void;
+
    int is_container;
+   int is_struct;
 };
+#define SLANG_CLASS_IS_SLSTRUCT(cl) ((cl)->is_struct != 0)
 
 extern int _pSLregister_types (void);
 extern SLang_Class_Type *_pSLclass_get_class (SLtype);
@@ -917,8 +962,8 @@ extern int _pSLclass_add_arith_unary_op (SLtype type,
 					int (*r)(int, SLtype, SLtype *));
 #endif
 
-extern int _pSLclass_get_unary_opcode (char *name);
-extern int _pSLclass_get_binary_opcode (char *name);
+extern int _pSLclass_get_unary_opcode (SLCONST char *name);
+extern int _pSLclass_get_binary_opcode (SLCONST char *name);
 extern int _pSLclass_is_same_obj (SLang_Object_Type *a, SLang_Object_Type *b);
 extern int _pSLclass_obj_eqs (SLang_Object_Type *a, SLang_Object_Type *b);
 
@@ -931,48 +976,58 @@ extern void _pSLang_set_arith_type (SLtype, unsigned char);
 #if SLANG_OPTIMIZE_FOR_SPEED
 extern SLclass_Type _pSLang_get_class_type (SLtype);
 extern void _pSLang_set_class_info (SLtype, SLang_Class_Type *);
+extern int _pSLarray_bin_op (SLang_Object_Type *, SLang_Object_Type *, int);
+extern int _pSLarray1d_push_elem (SLang_Array_Type *at, SLindex_Type idx);
 #endif
 extern int _pSLarith_bin_op (SLang_Object_Type *, SLang_Object_Type *, int);
 
+/* Does not perform any range checking.  It is up to the caller. 
+ * If *atp!=NULL, then it is an array of indices.
+ */
+extern int _pSLarray_pop_index (unsigned int num_elements, SLang_Array_Type **ind_atp, SLindex_Type *ind);
+
 extern int _pSLarray_add_bin_op (SLtype);
 extern int _pSLarray_push_elem_ref (void);
+extern int _pSLang_push_array (SLang_Array_Type *, int);   /* NULL not allowed */
 
 typedef struct
 {
-   char *name;
+   SLCONST char *name;
    SLang_Name_Type *next;
    char name_type;
 
    int unary_op;
 }
 SLang_Arith_Unary_Type;
-extern int _pSLadd_arith_unary_table (SLang_Arith_Unary_Type *tbl, char *);
+extern int _pSLadd_arith_unary_table (SLang_Arith_Unary_Type *tbl, SLFUTURE_CONST char *);
 
 typedef struct
 {
-   char *name;
+   SLCONST char *name;
    SLang_Name_Type *next;
    char name_type;
 
    int binary_op;
 }
 SLang_Arith_Binary_Type;
-extern int _pSLadd_arith_binary_table (SLang_Arith_Binary_Type *tbl, char *);
+extern int _pSLadd_arith_binary_table (SLang_Arith_Binary_Type *tbl, SLFUTURE_CONST char *);
 
 extern int _pSLang_do_binary_ab (int op, SLang_Object_Type *obja, SLang_Object_Type *objb);
 
 extern int _pSLang_call_funptr (SLang_Name_Type *);
-extern void _pSLset_double_format (char *);
-extern SLang_Name_Type *_pSLlocate_global_name (char *);
-extern SLang_Name_Type *_pSLlocate_name (char *);
+extern void _pSLset_double_format (SLCONST char *);
+extern SLCONST char *_pSLget_double_format (void);
+extern SLang_Name_Type *_pSLlocate_global_name (SLCONST char *);
+extern SLang_Name_Type *_pSLlocate_name (SLCONST char *);
 
-extern char *_pSLdefines[];
+extern SLFUTURE_CONST char *_pSLdefines[];
 
 #define SL_ERRNO_NOT_IMPLEMENTED	0x7FFF
 extern int _pSLerrno_errno;
 extern int _pSLerrno_init (void);
 
 extern int _pSLstdio_fdopen (char *, int, char *);
+extern void _pSLfclose_fdopen_fp (SLang_MMT_Type *);
 
 extern void _pSLstruct_pop_args (int *);
 extern void _pSLstruct_push_args (SLang_Array_Type *);
@@ -1020,6 +1075,32 @@ extern int _pSLassoc_inc_value (unsigned int, int val);
 extern int _pSLcompile_push_context (SLang_Load_Type *);
 extern int _pSLcompile_pop_context (void);
 extern int _pSLang_Auto_Declare_Globals;
+extern int _pSLang_Load_File_Verbose;
+
+typedef struct _pSLtoken_String_List_Type
+{
+   struct _pSLtoken_String_List_Type *next;
+   unsigned int len;
+   char buf[1];			       /* rest of bytes follow */
+}
+_pSLtoken_String_List_Type;
+
+typedef unsigned char _pSLtok_Type;
+typedef struct 
+{
+   _pSLtok_Type type;			       /* [b]string_token */
+   unsigned int num;
+   
+   _pSLtoken_String_List_Type *list;
+   union 
+     {
+	SLFUTURE_CONST char *s_val;		       /* (SLstring) concatenated strings */
+	SLang_BString_Type *b_val;     /* concatenated bstrings */
+     } v;
+   unsigned long hash;		       /* for v.s_val */
+   unsigned int len;		       /* length of v.?_val */
+}
+_pSLang_Multiline_String_Type;
 
 typedef struct _pSLang_Token_Type
 {
@@ -1038,29 +1119,42 @@ typedef struct _pSLang_Token_Type
 	 * We can avoid the issues associated with this by just storing 
 	 * floating point values as strings.
 	 */
-	char *s_val;		       /* Used for IDENT_TOKEN, DOUBLE_TOKEN, etc...  */
+	SLFUTURE_CONST char *s_val;		       /* Used for IDENT_TOKEN, DOUBLE_TOKEN, etc...  */
 	
 	SLang_BString_Type *b_val;
+	_pSLang_Multiline_String_Type *multistring_val;
      } v;
-   int free_sval_flag;
+   void (*free_val_func)(struct _pSLang_Token_Type *);
    unsigned int num_refs;
-   unsigned long hash;
+   unsigned long hash;		       /* hash for slstring, and length for _BSTRING */
+#define SLTOKEN_IS_NEGATIVE		0x001
+#define SLTOKEN_OVERFLOW_CHECKED	0x002
+#define SLTOKEN_IS_HEX			0x004
+#define SLTOKEN_TYPE_INTEGER		0x100
+#define SLTOKEN_TYPE_FLOAT		0x200
+#define SLTOKEN_TYPE_NUMBER		(SLTOKEN_TYPE_INTEGER|SLTOKEN_TYPE_FLOAT)
+   int flags;
+
 #if SLANG_HAS_DEBUG_CODE
    int line_number;
 #endif
    struct _pSLang_Token_Type *next;     /* used for token lists */
-   SLtype type;
+   _pSLtok_Type type;
 }
 _pSLang_Token_Type;
 
+/* return token type or EOF_TOKEN upon error */
+extern _pSLtok_Type _pSLtoken_init_slstring_token (_pSLang_Token_Type *, _pSLtok_Type, 
+						   SLFUTURE_CONST char *, unsigned int);
+
 extern void _pSLcompile (_pSLang_Token_Type *);
 extern void (*_pSLcompile_ptr)(_pSLang_Token_Type *);
 
 /* slmisc.c */
-extern char *_pSLskip_whitespace (char *s);
+extern char *_pSLskip_whitespace (SLCONST char *s);
 
 /* slospath.c */
-extern char *_pSLpath_find_file (char *, int);   /* slstring returned */
+extern char *_pSLpath_find_file (SLFUTURE_CONST char *, int);   /* slstring returned */
 
 /* Read but do not set this variable. */
 extern volatile int _pSLang_Error;
@@ -1123,8 +1217,8 @@ extern SLuchar_Type *_pSLinterp_encode_wchar (SLwchar_Type wch,
 #define IS_LVALUE_TOKEN (((t) <= DOT_TOKEN) && ((t) >= IDENT_TOKEN))
 #define DOT_METHOD_CALL_TOKEN		0x23
 
-#define ESC_STRING_TOKEN	0x24
-#define ESC_BSTRING_TOKEN	0x25
+#define ESC_STRING_TOKEN	0x24   /* only appears in slc file */
+#define ESC_BSTRING_TOKEN	0x25   /* only appears in slc file */
 
 /* Flags for struct fields */
 #define STATIC_TOKEN	0x26
@@ -1144,6 +1238,7 @@ extern SLuchar_Type *_pSLinterp_encode_wchar (SLwchar_Type wch,
 #define SEMICOLON_TOKEN	0x32
 #define COLON_TOKEN	0x33
 #define NAMESPACE_TOKEN	0x34
+#define QUESTION_TOKEN	0x35
 
 /* Operators */
 /* The order here must match the order in the Binop_Level table in slparse.c */
@@ -1255,6 +1350,8 @@ extern SLuchar_Type *_pSLinterp_encode_wchar (SLwchar_Type wch,
 #define THROW_TOKEN	0x89
 #define FINALLY_TOKEN	0x8a
 
+#define BREAK_N_TOKEN	0x8b
+#define CONT_N_TOKEN	0x8c
 #define THEN_TOKEN	0x8d
 
 #define STRUCT_WITH_ASSIGN_TOKEN	0x8e
@@ -1331,7 +1428,8 @@ extern SLuchar_Type *_pSLinterp_encode_wchar (SLwchar_Type wch,
 #define _INLINE_LIST_TOKEN		0xE4
 #define _INLINE_IMPLICIT_ARRAYN_TOKEN	0xE5   /* [a:b:#n] */
 
-#define ESC_STRING_DOLLAR_TOKEN		0xF0
+#define ESC_STRING_DOLLAR_TOKEN		0xF0   /* appears only in .slc files */
+#define MULTI_STRING_TOKEN		0xF1
 
 #define BOS_TOKEN			0xFA
 #define EOS_TOKEN			0xFB
@@ -1365,15 +1463,15 @@ typedef struct
 _pSLEval_Context;
 
 extern int _pSLget_token (_pSLang_Token_Type *);
-extern void _pSLparse_error (int, char *, _pSLang_Token_Type *, int);
+extern void _pSLparse_error (int, SLCONST char *, _pSLang_Token_Type *, int);
 extern void _pSLparse_start (SLang_Load_Type *);
 extern int _pSLget_rpn_token (_pSLang_Token_Type *);
 extern void _pSLcompile_byte_compiled (void);
 
-extern int (*_pSLprep_eval_hook) (char *);
+extern int (*_pSLprep_eval_hook) (SLFUTURE_CONST char *);
 
 extern int _pSLsecure_issetugid (void);
-extern char *_pSLsecure_getenv (char *);
+extern char *_pSLsecure_getenv (SLCONST char *);
 
 /* Error Handling */
 extern int _pSLang_init_exceptions (void);
@@ -1381,9 +1479,11 @@ extern int _pSLerr_init (void);
 extern void _pSLerr_deinit (void);
 extern int _pSLerr_suspend_messages (void);
 extern int _pSLerr_resume_messages (void);
-extern int _pSLerr_traceback_msg (char *, ...) SLATTRIBUTE_PRINTF(1,2);
-extern void _pSLerr_dump_msg (char *, ...) SLATTRIBUTE_PRINTF(1,2);
+extern int _pSLerr_traceback_msg (SLFUTURE_CONST char *, ...) SLATTRIBUTE_PRINTF(1,2);
+extern void _pSLerr_dump_msg (SLFUTURE_CONST char *, ...) SLATTRIBUTE_PRINTF(1,2);
 extern void _pSLerr_clear_error (int);
+extern void _pSLang_verror (int, SLCONST char *, ...) SLATTRIBUTE_PRINTF(2,3);
+extern int _pSLsnprintf (char *, unsigned int, SLFUTURE_CONST char *, ...) SLATTRIBUTE_PRINTF(3,4);
 
 typedef struct _pSLerr_Error_Queue_Type _pSLerr_Error_Queue_Type;
 extern _pSLerr_Error_Queue_Type *_pSLerr_new_error_queue (int);
@@ -1392,7 +1492,7 @@ extern void _pSLerr_delete_error_queue (_pSLerr_Error_Queue_Type *);
 
 extern void _pSLerr_print_message_queue (void);
 
-extern int _pSLerr_set_line_info (char *, int, char *);
+extern int _pSLerr_set_line_info (SLFUTURE_CONST char *, int, SLFUTURE_CONST char *);
 extern int _pSLang_pop_error_context (int);
 extern int _pSLang_push_error_context (void);
 extern void (*_pSLinterpreter_Error_Hook)(int);
@@ -1401,19 +1501,23 @@ extern int _pSLerr_get_last_error (void);
 extern int _pSLerr_pop_exception (int *);
 extern void _pSLerr_free_queued_messages (void);
 
-extern char *_pSLerr_get_error_from_queue (_pSLerr_Error_Queue_Type *);
+#define _SLERR_MSG_ERROR	1
+#define _SLERR_MSG_WARNING	2
+#define _SLERR_MSG_TRACEBACK	3
+extern char *_pSLerr_get_error_from_queue (_pSLerr_Error_Queue_Type *, int);
+
 extern int _pSLerr_throw (void);
-extern int (*_pSLerr_New_Exception_Hook)(char *name, char *desc, int error_code);
+extern int (*_pSLerr_New_Exception_Hook)(SLFUTURE_CONST char *name, SLFUTURE_CONST char *desc, int error_code);
 
 #if SLANG_HAS_DEBUGGER_SUPPORT
 typedef struct
 {
    char **locals;
    unsigned int nlocals;
-   char *file;
+   SLCONST char *file;
    unsigned int line;
-   char *function;
-   char *ns;
+   SLCONST char *function;
+   SLCONST char *ns;
 }
 _pSLang_Frame_Info_Type;
 
@@ -1429,6 +1533,25 @@ extern int _pSLmath_isnan (double x);
 extern int _pSLmath_isinf (double x);
 extern double _pSLang_NaN;
 extern double _pSLang_Inf;
+
+# ifdef HAVE_LOG1P
+#  define LOG1P_FUNC log1p
+# else
+extern double _pSLmath_log1p (double);
+#  define LOG1P_FUNC _pSLmath_log1p
+# endif
+
+# ifdef HAVE_EXPM1
+#  define EXPM1_FUNC expm1
+# else
+extern double _pSLmath_expm1 (double);
+#  define EXPM1_FUNC _pSLmath_expm1
+# endif
+
+# if SLANG_HAS_COMPLEX
+extern double *_pSLcomplex_expm1 (double *f, double *z);
+extern double *_pSLcomplex_log1p (double *f, double *z);
+# endif
 #endif
 
 #if SLANG_HAS_SIGNALS
