@@ -2,7 +2,7 @@
 #define DAVIS_SLANG_H_
 /* -*- mode: C; mode: fold; -*- */
 /*
-Copyright (C) 2004-2008 John E. Davis
+Copyright (C) 2004-2009 John E. Davis
 
 This file is part of the S-Lang Library.
 
@@ -22,8 +22,8 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 USA.  
 */
 
-#define SLANG_VERSION 20104
-#define SLANG_VERSION_STRING "2.1.4"
+#define SLANG_VERSION 20200
+#define SLANG_VERSION_STRING "2.2.0"
 /* #ifdef __DATE__ */
 /* # define SLANG_VERSION_STRING SLANG_VERSION_STRING0 " " __DATE__ */
 /* #else */
@@ -405,7 +405,7 @@ SLuchar_Type *SLuchar_apply_char_map (SLwchar_Map_Type *map, SLuchar_Type *str);
 /*{{{ Interpreter Typedefs */
 
 typedef unsigned int SLtype;
-/* typedef unsigned short SLtype; */
+#define SLANG_SLTYPE_INT_TYPE SLANG_UINT_TYPE
 
 typedef struct _pSLang_Name_Type
 {
@@ -774,6 +774,8 @@ SL_EXTERN int SLclass_typecast (SLtype, int, int);
 #define SLMATH_FLOOR	23
 #define SLMATH_CEIL	24
 #define SLMATH_ROUND	25
+#define SLMATH_EXPM1	26
+#define SLMATH_LOG1P	27
 
 SL_EXTERN int SLclass_add_unary_op (SLtype,
 				 int (*) (int,
@@ -869,10 +871,11 @@ typedef struct _pSLang_Array_Type
     * value at the specified index location.
     */
    unsigned int flags;
-#define SLARR_DATA_VALUE_IS_READ_ONLY		1
-#define SLARR_DATA_VALUE_IS_POINTER		2
-#define SLARR_DATA_VALUE_IS_RANGE		4
-#define SLARR_DATA_VALUE_IS_INTRINSIC		8
+#define SLARR_DATA_VALUE_IS_READ_ONLY		0x0001
+#define SLARR_DATA_VALUE_IS_POINTER		0x0002
+#define SLARR_DATA_VALUE_IS_RANGE		0x0004
+#define SLARR_DATA_VALUE_IS_INTRINSIC		0x0008
+#define SLARR_DERIVED_FROM_SCALAR		0x0100
    SLang_Class_Type *cl;
    unsigned int num_refs;
    void (*free_fun)_PROTO((struct _pSLang_Array_Type *));
@@ -880,6 +883,11 @@ typedef struct _pSLang_Array_Type
 }
 SLang_Array_Type;
 
+extern int _pSLarray_convert_to_array (VOID_STAR cd, 
+				       int (*get_type)(VOID_STAR, SLuindex_Type, SLtype *),
+				       int (*push)(VOID_STAR, SLuindex_Type),
+				       SLuindex_Type num_objects, SLtype type);
+
 SL_EXTERN int SLang_pop_array_of_type (SLang_Array_Type **atp, SLtype type);
 SL_EXTERN int SLang_pop_array (SLang_Array_Type **atp, int convert_scalar);
 SL_EXTERN int SLang_push_array (SLang_Array_Type *at, int do_free);
@@ -1998,6 +2006,7 @@ SL_EXTERN SLSig_Fun_Type *SLsignal_intr (int, SLSig_Fun_Type *);
 SL_EXTERN int SLsig_block_signals (void);
 SL_EXTERN int SLsig_unblock_signals (void);
 SL_EXTERN int SLsystem (SLFUTURE_CONST char *);
+SL_EXTERN int SLsystem_intr (SLFUTURE_CONST char *);
 
 /* Make a signal off-limits to the interpreter */
 SL_EXTERN int SLsig_forbid_signal (int);
