@@ -1,6 +1,6 @@
 /* User defined objects */
 /*
-Copyright (C) 2004, 2005, 2006, 2007, 2008 John E. Davis
+Copyright (C) 2004-2009 John E. Davis
 
 This file is part of the S-Lang Library.
 
@@ -21,6 +21,7 @@ USA.
 */
 
 #include "slinclud.h"
+#include <errno.h>
 
 /* #define SL_APP_WANTS_FOREACH */
 #include "slang.h"
@@ -299,21 +300,92 @@ static int scalar_fread (SLtype type, FILE *fp, VOID_STAR ptr,
 			 unsigned int desired, unsigned int *actual)
 {
    unsigned int n;
+   char *buf = (char *)ptr;
+   size_t desired_bytes, actual_bytes;
+   size_t size = _pSLclass_get_class (type)->cl_sizeof_type;
 
-   n = fread ((char *) ptr, _pSLclass_get_class (type)->cl_sizeof_type,
-	      desired, fp);
-   *actual = n;
+   desired_bytes = size * desired;
+   actual_bytes = 0;
+
+   while (desired_bytes)
+     {
+	int e;
+
+	errno = 0;
+	n = fread (buf, 1, desired_bytes, fp);
+
+	actual_bytes += n;
+	if (n == desired_bytes)
+	  break;
+
+	e = errno;
+	desired_bytes -= n;
+	buf += n;
+
+	clearerr (fp);
+#ifdef EINTR
+	if ((e == EINTR)
+	    && (0 == SLang_handle_interrupt ()))
+	  continue;
+#endif
+	_pSLerrno_errno = e;
+	break;
+     }
+
+   if (actual_bytes % size)
+     {
+	/* Sigh.  We failed to read a full object. */
+     }
+   *actual = actual_bytes / size;
    return 0;
 }
 
 static int scalar_fwrite (SLtype type, FILE *fp, VOID_STAR ptr,
-			  unsigned int desired, unsigned int *actual)
+			 unsigned int desired, unsigned int *actual)
 {
    unsigned int n;
+   char *buf = (char *)ptr;
+   size_t desired_bytes, actual_bytes;
+   size_t size = _pSLclass_get_class (type)->cl_sizeof_type;
 
-   n = fwrite ((char *) ptr, _pSLclass_get_class (type)->cl_sizeof_type,
-	       desired, fp);
-   *actual = n;
+   desired_bytes = size * desired;
+   actual_bytes = 0;
+
+   while (desired_bytes)
+     {
+	int e;
+
+	errno = 0;
+	n = fwrite (buf, 1, desired_bytes, fp);
+
+	actual_bytes += n;
+	if (n == desired_bytes)
+	  break;
+
+	e = errno;
+	desired_bytes -= n;
+	buf += n;
+
+	clearerr (fp);
+#ifdef EINTR
+	if ((e == EINTR)
+	    && (0 == SLang_handle_interrupt ()))
+	  continue;
+#endif
+	_pSLerrno_errno = e;
+
+	/* Apparantly, the write can be interrupted returning a short item
+	 * count but not set errno.
+	 */
+	if (n == 0)
+	  break;
+     }
+
+   if (actual_bytes % size)
+     {
+	/* Sigh.  We failed to write out a full object. */
+     }
+   *actual = actual_bytes / size;
    return 0;
 }
 
