@@ -1,5 +1,5 @@
 /*
-Copyright (C) 2004-2008 John E. Davis
+Copyright (C) 2004-2009 John E. Davis
 
 This file is part of the S-Lang Library.
 
@@ -207,6 +207,9 @@ USA.
 #undef HAVE_UNAME
 
 #undef HAVE_POPEN
+#if defined(__MINGW32__)
+# define HAVE_POPEN 1
+#endif
 
 /* #define HAVE_ROUND	1 */
 #if !defined(__os2__) && !defined(__WATCOMC__)
@@ -219,6 +222,10 @@ USA.
 #endif
 #define HAVE_HYPOT	1
 #define HAVE_ATAN2	1
+#define HAVE_FREXP	1
+#undef HAVE_FREXPF
+#define HAVE_LDEXP	1
+#undef HAVE_LDEXPF
 
 #if defined(__WIN32__)
 # define isnan _isnan
@@ -243,23 +250,24 @@ USA.
 #ifdef _MSC_VER
 # define HAVE_VSNPRINTF 1
 # define HAVE_SNPRINTF 1
-/* Grrr, Visual Studio has vnsprintf, but with an '_'. */
-# define vsnprintf _vsnprintf
+# if _MSC_VER < 1400
+/* Grrr, older versions of Visual Studio have vnsprintf, but with an '_'. */
+#  define vsnprintf _vsnprintf
+# endif
 #endif
 
-#if defined(__unix__) || defined(__DECC) || defined(__BORLAND_V5__)
-#else
-# define mode_t int
-# if !defined(_MSC_VER) && !defined(__WATCOMC__)
-#  define pid_t int
-#  define uid_t int
-#  define gid_t int
-#  define off_t int
-# endif
-/* # define size_t unsigned int */
-#endif
+/* Some compilers may not have definitions for these. */
+/* #define mode_t int */
+/* #define pid_t int */
+/* #define uid_t int */
+/* #define gid_t int */
+/* #define off_t int */
+/* #define size_t unsigned int */
+
 #undef HAVE_LONG_LONG
 
+#define HAVE_SETVBUF
+
 #define SIZEOF_OFF_T	4
 
 #if defined (__EMX__) || defined(__BORLANDC__) || defined(__IBMC__) || defined(_MSC_VER) || defined(__WATCOMC__)
@@ -314,6 +322,8 @@ USA.
 # define HAVE_STRTOLL	1
 #endif
 
+#undef HAVE_FORK
+#undef HAVE_WAITPID
 
 /* Files used by the slang-readline interface.  The interface first tries
  * to read SLRLINE_USER_INIT_FILE from $HOME, and if that does not exist, it
