@@ -1,5 +1,5 @@
 /*
-Copyright (C) 2004, 2005, 2006, 2007, 2008 John E. Davis
+Copyright (C) 2004-2009 John E. Davis
 
 This file is part of the S-Lang Library.
 
@@ -2197,7 +2197,8 @@ static char *fixup_tgetstr (char *what)
 	if (*w == 0) break;
 	w++;
 	wsave = w1;
-	while ((*w1++ = *w++) != 0);
+	while ((*w1++ = *w++) != 0)
+	  ;
 	w = wsave;
      }
 
@@ -2233,10 +2234,9 @@ The integrity of this program has been violated.\n");
     * this more general.
     */
    /* FIXME: Priority=low; */
-   if (0 == strcmp (cap, "ac"))
-     return s;
+   if (0 != strcmp (cap, "ac"))
+     s = fixup_tgetstr (s);
 
-   s = fixup_tgetstr (s);
 #ifdef USE_TERMCAP
    if ((s >= area_buf) && (s < area_buf + sizeof(area_buf)))
      {
