@@ -1,6 +1,6 @@
 /* sin, cos, etc, for S-Lang */
 /*
-Copyright (C) 2004, 2005, 2006, 2007, 2008 John E. Davis
+Copyright (C) 2004-2009 John E. Davis
 
 This file is part of the S-Lang Library.
 
@@ -65,6 +65,9 @@ static void math_floating_point_exception (int sig)
 
 double SLmath_hypot (double x, double y)
 {
+#ifdef HAVE_HYPOT
+   return hypot (x,y);
+#else
    double fr, fi, ratio;
 
    fr = fabs(x);
@@ -83,25 +86,7 @@ double SLmath_hypot (double x, double y)
      }
 
    return x;
-}
-
-/* usage here is a1 a2 ... an n x ==> a1x^n + a2 x ^(n - 1) + ... + an */
-static double math_poly (void)
-{
-   int n;
-   double xn = 1.0, sum = 0.0;
-   double an, x;
-
-   if ((SLang_pop_double(&x))
-       || (SLang_pop_integer(&n))) return(0.0);
-
-   while (n-- > 0)
-     {
-	if (SLang_pop_double(&an)) break;
-	sum += an * xn;
-	xn = xn * x;
-     }
-   return (double) sum;
+#endif
 }
 
 static int double_math_op_result (int op, SLtype a, SLtype *b)
@@ -184,31 +169,112 @@ static double my_round (double x)
 }
 #endif
 
+#ifndef HAVE_LOG1P
+double _pSLmath_log1p (double x)
+{
+   /*
+    * log(1+x) ~= x for small x
+    * Suppose the error in 1+x is e.  That is, 1+x produces 1+x+e.
+    * Then log(1+x) will produce log(1+x+e).  The error is
+    *   df = log(1+x+e)-log(1+x)
+    *      = log (1 + e/(1+x))
+    *      ~= e/(1+x)
+    * Here e = (1+x) - (1+x)
+    *        = ((1+x)-1) - x
+    * So compute:
+    *   log1p(x) = log(1+x) - (((1+x)-1)-x)/(1+x)
+    */
+   volatile double xp1 = 1.0 + x;
+   volatile double xx = xp1-1.0;
+   return log(xp1) - (xx-x)/xp1;
+}
+#endif
+
+#ifndef HAVE_EXPM1
+double _pSLmath_expm1 (double x)
+{
+   /* f = exp(x)-1
+    *    = (exp(x)-1) + x - x
+    *    = (exp(x)-(1+x)) + x
+    * 
+    * While the above works, it is not as accurate as the algroithm below,
+    * which is due to Kahan (yes, that Kahan).
+    */
+   volatile double u = exp(x);
+   if (u == 1.0)
+     return x;
+   if (u - 1.0 == -1.0)
+     return -1.0;
+   return (u-1.0)*x/log(u);
+}
+#endif
+
 #ifdef HAVE_ASINH
 # define ASINH_FUN	asinh
 #else
 # define ASINH_FUN	my_asinh
 static double my_asinh (double x)
 {
-   return log (x + sqrt (x*x + 1));
+   /* f(x) = log(x + sqrt(x^2+1))
+    * Note: f(x)+f(-x) 
+    *          = log(x+sqrt(x^2+1))+log(-x+sqrt(x^2+1))
+    *          = log(x^2+1 - x^2)
+    *          = 0
+    * Thus, f(-x)=-f(x) as required.
+    * 
+    * So consider x>=0.
+    * For large x
+    *   = log(2x + sqrt(x^2+1)-x)
+    *   = log(2x + ((x^2+1)-x^2)/(x+sqrt(x^2+1)))
+    *   = log(2x + 1/(x+sqrt(x^2+1)))
+    *   = log(2x + 1/x/(1+sqrt(1+1/x^2)))
+    * For small x
+    *   = log(1 + x + sqrt(1+x^2)-1)
+    *   = log(1 + x + (1+x^2-1)/(sqrt(1+x^2)+1))
+    *   = log(1 + x + x^2/(1+sqrt(1+x^2)))
+    */
+   double s = 1.0;
+   double x2 = x*x;
+
+   if (x < 0.0)
+     {
+	s = -1.0;
+	x = -x;
+     }
+
+   if (x > 1.0)
+     return s*log (2.0*x + 1.0/x/(1.0+sqrt(1.0+1.0/x2)));
+   
+   return LOG1P_FUNC(x+x2/(1.0+sqrt(1.0+x2)));
 }
 #endif
+
 #ifdef HAVE_ACOSH
 # define ACOSH_FUN	acosh
 #else
 # define ACOSH_FUN	my_acosh
 static double my_acosh (double x)
 {
-   return log (x + sqrt(x*x - 1));     /* x >= 1 */
+   /* log (x + sqrt(x^2 - 1);     x >= 1
+    *  = log(2*x + sqrt(x^2-1)-x)
+    *  = log(2*x + (x^2-1-x^2)/(x+sqrt(x^2-1)))
+    *  = log(2*x - 1/x/(1+sqrt(1-1/x^2)))
+    */
+   return log(2.0*x - 1.0/x/(1.0+sqrt(1.0-1/x/x)));
 }
 #endif
 #ifdef HAVE_ATANH
 # define ATANH_FUN	atanh
 #else
 # define ATANH_FUN	my_atanh
+
 static double my_atanh (double x)
 {
-   return 0.5 * log ((1.0 + x)/(1.0 - x)); /* 0 <= x^2 < 1 */
+   /* 0.5 * log ((1.0 + x)/(1.0 - x));  (0 <= x^2 < 1)
+    * = 0.5*log((1-x+2x)/(1-x)) 
+    * = 0.5*log(1 + 2x/(1-x))
+    */
+   return 0.5 * LOG1P_FUNC(2.0*x/(1-x));
 }
 #endif
 
@@ -311,6 +377,14 @@ static int double_math_op (int op,
       case SLMATH_ROUND:
 	for (i = 0; i < na; i++) b[i] = ROUND_FUN(a[i]);
 	break;
+	
+      case SLMATH_EXPM1:
+	for (i = 0; i < na; i++) b[i] = EXPM1_FUNC(a[i]);
+	break;
+
+      case SLMATH_LOG1P:
+	for (i = 0; i < na; i++) b[i] = LOG1P_FUNC(a[i]);
+	break;
      }
    
    return 1;
@@ -416,6 +490,14 @@ static int float_math_op (int op,
       case SLMATH_ROUND:
 	for (i = 0; i < na; i++) b[i] = ROUND_FUN(a[i]);
 	break;
+
+      case SLMATH_EXPM1:
+	for (i = 0; i < na; i++) b[i] = EXPM1_FUNC(a[i]);
+	break;
+
+      case SLMATH_LOG1P:
+	for (i = 0; i < na; i++) b[i] = LOG1P_FUNC(a[i]);
+	break;
      }
 
    return 1;
@@ -608,6 +690,22 @@ static int generic_math_op (int op,
 	     a += da;
 	  }
 	break;
+
+      case SLMATH_EXPM1:
+	for (i = 0; i < na; i++) 
+	  {
+	     b[i] = EXPM1_FUNC (to_double ((VOID_STAR) a));
+	     a += da;
+	  }
+	break;
+
+      case SLMATH_LOG1P:
+	for (i = 0; i < na; i++) 
+	  {
+	     b[i] = LOG1P_FUNC (to_double ((VOID_STAR) a));
+	     a += da;
+	  }
+	break;
      }
    
    return 1;
@@ -762,7 +860,7 @@ typedef struct
    double *dptr;
    char *cptr;
    unsigned int inc;		       /* inc = 0, if at==NULL */
-   unsigned int num;
+   SLuindex_Type num;
 }
 Array_Or_Scalar_Type;
 
@@ -1018,7 +1116,209 @@ static int do_binary_function (double (*f)(double, double))
 
 static void hypot_fun (void)
 {
-   (void) do_binary_function (SLmath_hypot);
+   Array_Or_Scalar_Type ast;
+   unsigned int num;
+
+   if (SLang_Num_Function_Args == 2)
+     {
+	(void) do_binary_function (SLmath_hypot);
+	return;
+     }
+   if (-1 == pop_array_or_scalar (&ast))
+     return;
+
+   if (0 == (num = ast.num))
+     {
+	SLang_verror (SL_InvalidParm_Error, "An empty array was passed to hypot");
+	free_array_or_scalar (&ast);
+	return;
+     }
+
+   if (ast.is_float)
+     {
+	float *f = ast.fptr;
+	double sum, esum, max;
+	unsigned int i;
+
+	max = fabs((double)*f);
+	
+	for (i = 1; i < num; i++)
+	  {
+	     double max1 = fabs((double)f[i]);
+	     if (max1 > max) max  = max1;
+	  }
+	sum = 0.0; esum = 0.0;
+	if (max > 0.0) for (i = 0; i < num; i++)
+	  {
+	     double term = f[i]/max;
+	     double new_sum;
+	     
+	     term = term*term;
+	     new_sum = sum + term;
+	     esum += term - (new_sum-sum);
+	     sum = new_sum;
+	  }
+	(void) SLang_push_float ((float) max*sqrt(sum+esum));
+     }
+   else
+     {
+	unsigned int i;
+	double *d = ast.dptr;
+	double sum, esum, max;
+	max = fabs(*d);
+	
+	for (i = 1; i < num; i++)
+	  {
+	     double max1 = fabs(d[i]);
+	     if (max1 > max) max  = max1;
+	  }
+	sum = 0.0; esum = 0.0;
+	if (max > 0.0) for (i = 0; i < num; i++)
+	  {
+	     double term = d[i]/max;
+	     double new_sum;
+	     
+	     term = term*term;
+	     new_sum = sum + term;
+	     esum += term - (new_sum-sum);
+	     sum = new_sum;
+	  }
+	(void) SLang_push_double (max*sqrt(sum+esum));
+     }
+   free_array_or_scalar (&ast);
+}
+
+static void math_poly (void)
+{
+   Array_Or_Scalar_Type ast;
+   SLang_Array_Type *coeff_at, *y_at;
+   SLuindex_Type i, k, n;
+   SLuindex_Type num;
+   double *a;
+   double x, y;
+   int use_factorial = 0;
+
+   switch (SLang_Num_Function_Args)
+     {
+      case 3:
+	if (-1 == SLang_pop_int (&use_factorial))
+	  return;
+	/* drop */
+      case 2:
+	break;
+	
+      default:
+	SLang_verror (SL_Usage_Error, "\
+Usage: y = polynom([a0,a1,...], x [,use_factorial_form])");
+	return;
+     }
+
+   if (-1 == pop_array_or_scalar (&ast))
+     return;
+
+   if (-1 == SLang_pop_array_of_type (&coeff_at, SLANG_DOUBLE_TYPE))
+     {
+	free_array_or_scalar (&ast);
+	return;
+     }
+   a = (double *) coeff_at->data;
+   n = coeff_at->num_elements;
+
+   if (ast.inc == 0)
+     {
+	if (ast.is_float)
+	  x = (double) ast.f;
+	else
+	  x = ast.d;
+
+	y = 0.0;
+	k = n;
+
+	if (use_factorial)
+	  {
+	     while (k != 0)
+	       {
+		  y = a[k-1] + (x/k)*y;
+		  k--;
+	       }
+	  }
+	else while (k != 0)
+	  {
+	     k--;
+	     y = a[k] + x*y;
+	  }
+	
+	if (ast.is_float)
+	  (void) SLang_push_float ((float) y);
+	else
+	  (void) SLang_push_double (y);
+
+	goto free_and_return;
+     }
+   
+   if (NULL == (y_at = create_from_tmp_array (ast.at, NULL, ast.at->data_type)))
+     goto free_and_return;
+
+   num = ast.num;
+
+   if (ast.is_float)
+     {
+	float *f = ast.fptr;
+	float *yf = (float *)y_at->data;
+
+	for (i = 0; i < num; i++)
+	  {
+	     x = (double) f[i];
+	     y = 0.0;
+	     k = n;
+	     if (use_factorial)
+	       {
+		  while (k != 0)
+		    {
+		       y = a[k-1] + (x/k)*y;
+		       k--;
+		    }
+	       }
+	     else while (k != 0)
+	       {
+		  k--;
+		  y = a[k] + x*y;
+	       }
+	     yf[i] = (float) y;
+	  }
+     }
+   else
+     {
+	double *d = ast.dptr;
+	double *yd = (double *)y_at->data;
+
+	for (i = 0; i < num; i++)
+	  {
+	     x = d[i];
+	     y = 0.0;
+	     k = n;
+	     if (use_factorial)
+	       {
+		  while (k != 0)
+		    {
+		       y = a[k-1] + (x/k)*y;
+		       k--;
+		    }
+	       }
+	     else while (k != 0)
+	       {
+		  k--;
+		  y = a[k] + x*y;
+	       }
+	     yd[i] = y;
+	  }
+     }
+   
+   (void) SLang_push_array (y_at, 1);
+   /* drop */
+free_and_return:
+   free_array_or_scalar (&ast);
+   SLang_free_array (coeff_at);
 }
 
 static void atan2_fun (void)
@@ -1383,6 +1683,231 @@ static void nint_intrin (void)
    SLang_free_array (at);
 }
 
+#ifdef HAVE_FREXP
+# define FREXP_FUN(x,e) frexp(x,e)
+# ifdef HAVE_FREXPF
+#  define FREXPF_FUN(x,e) frexpf(x,e)
+# else
+#  define FREXPF_FUN(x,e) (float)FREXP_FUN(x,e)
+# endif
+
+static void frexp_intrin (void)
+{
+   double d;
+   float f;
+   int e, *ep;
+   SLuindex_Type i, imax;
+   SLang_Array_Type *at, *bt, *et;
+
+   switch (SLang_peek_at_stack ())
+     {
+      case SLANG_ARRAY_TYPE:
+	break;
+
+      case SLANG_FLOAT_TYPE:
+	if (-1 == SLang_pop_float (&f))
+	  return;
+	f = FREXPF_FUN(f, &e);
+	(void) SLang_push_float (f);
+	(void) SLang_push_int (e);
+	return;
+	
+      default:
+      case SLANG_DOUBLE_TYPE:
+	if (-1 == SLang_pop_double (&d))
+	  return;
+	d = FREXP_FUN(d, &e);
+	(void) SLang_push_double (d);
+	(void) SLang_push_int (e);
+	return;
+     }
+   
+   switch (SLang_peek_at_stack1 ())
+     {
+      case SLANG_FLOAT_TYPE:
+	if (-1 == SLang_pop_array_of_type (&at, SLANG_FLOAT_TYPE))
+	  return;
+	break;
+	
+      default:
+      case SLANG_DOUBLE_TYPE:
+	if (-1 == SLang_pop_array_of_type (&at, SLANG_DOUBLE_TYPE))
+	  return;
+	break;
+     }
+   
+   if (NULL == (bt = SLang_create_array1 (at->data_type, 0, NULL, at->dims, at->num_dims, 1)))
+     {
+	SLang_free_array (at);
+	return;
+     }
+   if (NULL == (et = SLang_create_array1 (SLANG_INT_TYPE, 0, NULL, at->dims, at->num_dims, 1)))
+     {
+	SLang_free_array (at);
+	SLang_free_array (bt);
+	return;
+     }
+   
+   imax = at->num_elements;
+   ep = (int *)et->data;
+
+   if (at->data_type == SLANG_DOUBLE_TYPE)
+     {
+	double *a = (double *)at->data;
+	double *b = (double *)bt->data;
+	for (i = 0; i < imax; i++)
+	  {
+	     b[i] = FREXP_FUN(a[i], ep+i);
+	  }
+     }
+   else
+     {
+	float *a = (float *)at->data;
+	float *b = (float *)bt->data;
+	for (i = 0; i < imax; i++)
+	  {
+	     b[i] = FREXPF_FUN(a[i], ep+i);
+	  }
+     }
+   
+   (void) SLang_push_array (bt, 0);
+   (void) SLang_push_array (et, 0);
+   SLang_free_array (et);
+   SLang_free_array (bt);
+   SLang_free_array (at);
+}
+#endif				       /* HAVE_FREXP */
+
+#ifdef HAVE_LDEXP
+# define LDEXP_FUN(a,b) ldexp(a,b)
+# ifdef HAVE_LDEXPF
+#  define LDEXPF_FUN(a,b) ldexpf(a,b)
+# else
+#  define LDEXPF_FUN(a,b) (float)LDEXP_FUN(a,b)
+# endif
+static void ldexp_intrin (void)
+{
+   Array_Or_Scalar_Type ast;
+   SLang_Array_Type *e_at = NULL, *c_at;
+   int *e_ptr, e_data;
+   SLuindex_Type i, imax;
+
+   if (SLang_peek_at_stack () == SLANG_ARRAY_TYPE)
+     {
+	if (-1 == SLang_pop_array_of_type (&e_at, SLANG_INT_TYPE))
+	  return;
+	e_ptr = (int *)e_at->data;
+     }
+   else 
+     {
+	if (-1 == SLang_pop_int (&e_data))
+	  return;
+	e_ptr = &e_data;
+     }
+
+   if (-1 == pop_array_or_scalar (&ast))
+     {
+	if (e_at != NULL)
+	  SLang_free_array (e_at);
+	return;
+     }
+
+   if ((e_at == NULL) && (ast.at == NULL))
+     {
+	/* Scalar case */
+	if (ast.is_float)
+	  (void) SLang_push_float (LDEXPF_FUN(ast.f, *e_ptr));
+	else
+	  (void) SLang_push_double (LDEXP_FUN(ast.d, *e_ptr));
+	/* free_array_or_scalar (&ast);   Nothing to free */
+	return;
+     }
+
+   if (NULL == (c_at = create_from_tmp_array (ast.at, e_at, ast.is_float ? SLANG_FLOAT_TYPE : SLANG_DOUBLE_TYPE)))
+     {
+	free_array_or_scalar (&ast);
+	SLang_free_array (e_at);
+	return;
+     }
+
+   if (e_at == NULL)
+     {
+	int e = *e_ptr;
+	imax = ast.num;
+
+	if (ast.is_float)
+	  {
+	     float *c, *a;
+	     a = ast.fptr;
+	     c = (float *)c_at->data;
+	     for (i = 0; i < imax; i++)
+	       c[i] = LDEXPF_FUN(a[i], e);
+	  }
+	else
+	  {
+	     double *c, *a;
+	     a = ast.dptr;
+	     c = (double *)c_at->data;
+	     for (i = 0; i < imax; i++)
+	       c[i] = LDEXP_FUN(a[i], e);
+	  }
+	goto push_free_and_return;
+     }
+
+   if (ast.at == NULL)
+     {
+	imax = e_at->num_elements;
+
+	if (ast.is_float)
+	  {
+	     float *c, a;
+	     a = ast.f;
+	     c = (float *)c_at->data;
+	     for (i = 0; i < imax; i++)
+	       c[i] = LDEXPF_FUN(a, e_ptr[i]);
+	  }
+	else
+	  {
+	     double *c, a;
+	     a = ast.d;
+	     c = (double *)c_at->data;
+	     for (i = 0; i < imax; i++)
+	       c[i] = LDEXP_FUN(a, e_ptr[i]);
+	  }
+	goto push_free_and_return;
+     }
+   
+   if (e_at->num_elements != ast.num)
+     {
+	SLang_verror (SL_TypeMismatch_Error, "ldexp: Array sizes do not match");
+	goto free_and_return;
+     }
+
+   imax = ast.num;
+   if (ast.is_float)
+     {
+	float *c = (float *)c_at->data;
+	float *a = ast.fptr;
+	for (i = 0; i < imax; i++)
+	  c[i] = LDEXPF_FUN(a[i], e_ptr[i]);
+     }
+   else
+     {
+	double *c = (double *)c_at->data;
+	double *a = ast.dptr;
+	for (i = 0; i < imax; i++)
+	  c[i] = LDEXP_FUN(a[i], e_ptr[i]);
+     }
+   /* drop */
+push_free_and_return:
+   (void) SLang_push_array (c_at, 0);
+   /* drop */
+free_and_return:
+   if (e_at != NULL) SLang_free_array (e_at);
+   SLang_free_array (c_at);
+   free_array_or_scalar (&ast);
+}
+#endif				       /* HAVE_LDEXPF */
 
 static void fpu_clear_except_bits (void)
 {
@@ -1424,6 +1949,8 @@ static SLang_Math_Unary_Type SLmath_Table [] =
    MAKE_MATH_UNARY("floor", SLMATH_FLOOR),
    MAKE_MATH_UNARY("ceil", SLMATH_CEIL),
    MAKE_MATH_UNARY("round", SLMATH_ROUND),
+   MAKE_MATH_UNARY("expm1", SLMATH_EXPM1),
+   MAKE_MATH_UNARY("log1p", SLMATH_LOG1P),
    
 #if SLANG_HAS_COMPLEX
    MAKE_MATH_UNARY("Real", SLMATH_REAL),
@@ -1436,7 +1963,7 @@ static SLang_Math_Unary_Type SLmath_Table [] =
 static SLang_Intrin_Fun_Type SLang_Math_Table [] =
 {
    MAKE_INTRINSIC_0("nint", nint_intrin, SLANG_VOID_TYPE),
-   MAKE_INTRINSIC_0("polynom", math_poly, SLANG_DOUBLE_TYPE),
+   MAKE_INTRINSIC_0("polynom", math_poly, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_0("hypot", hypot_fun, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_0("atan2", atan2_fun, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_0("_min", min_fun, SLANG_VOID_TYPE),
@@ -1448,6 +1975,12 @@ static SLang_Intrin_Fun_Type SLang_Math_Table [] =
    MAKE_INTRINSIC_0("fgteqs", fgeqs_fun, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_0("fpu_clear_except_bits", fpu_clear_except_bits, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_1("fpu_test_except_bits", fpu_test_except_bits, _pSLANG_LONG_TYPE, _pSLANG_LONG_TYPE),
+#ifdef HAVE_FREXP
+   MAKE_INTRINSIC_0("frexp", frexp_intrin, SLANG_VOID_TYPE),
+#endif
+#ifdef HAVE_LDEXP
+   MAKE_INTRINSIC_0("ldexp", ldexp_intrin, SLANG_VOID_TYPE),
+#endif
    SLANG_END_INTRIN_FUN_TABLE
 };
 
