@@ -1,6 +1,6 @@
 /* file stdio intrinsics for S-Lang */
 /*
-Copyright (C) 2004, 2005, 2006, 2007, 2008 John E. Davis
+Copyright (C) 2004-2009 John E. Davis
 
 This file is part of the S-Lang Library.
 
@@ -81,6 +81,9 @@ typedef struct
 #define SL_FDOPEN	0x2000
 #define SL_PIPE		0x4000
 #define SL_INUSE	0x8000
+   
+   char *buf;
+   unsigned int buflen;
 }
 SL_File_Table_Type;
 
@@ -117,30 +120,54 @@ static int handle_errno (int e)
 	  }
      }
 #endif
-   _pSLerrno_errno = errno;
+   _pSLerrno_errno = e;
    return 0;
 }
 
 /* The man page for fputs indicates that it returns a non-negative integer, or
  * EOF.  It says nothing about how many bytes were sucessfully written.  Sigh.
+ * So, use fwrite instead.
  */
 static int signal_safe_fputs (char *buf, FILE *fp)
 {
+   unsigned int len;
+   unsigned int num_written;
+
+   len = strlen (buf);
+   num_written = 0;
    errno = 0;
-   if (EOF != fputs (buf, fp))
-     return 0;
-   _pSLerrno_errno = errno;
-   return -1;
+   while (num_written < len)
+     {
+	unsigned int n = len - num_written;
+	unsigned int dn = fwrite (buf + num_written, 1, n, fp);
+
+	num_written += dn;
+	if (dn < n)
+	  {
+	     int e = errno;
+
+	     _pSLerrno_errno = e;
+	     clearerr (fp);
+	     if (n == 0)
+	       {
+		  if (0 == handle_errno (e))
+		    break;
+	       }
+	  }
+     }
+   return num_written;
 }
 
 static int signal_safe_fgets (char *buf, unsigned int buflen, FILE *fp)
 {
    errno = 0;
 
-   if (NULL != fgets (buf, buflen, fp))
-     return 0;
-   _pSLerrno_errno = errno;
-   return -1;
+   while (NULL == fgets (buf, buflen, fp))
+     {
+	if (0 == handle_errno (errno))
+	  return -1;
+     }
+   return 0;
 }
 
 
@@ -236,14 +263,26 @@ static int open_file_type (char *file, int fd, char *mode,
 
 static FILE *fopen_fun (char *f, char *m)
 {
-   return fopen (f, m);
+   FILE *fp;
+
+   errno = 0;
+   while (NULL == (fp = fopen (f, m)))
+     {
+	if (0 == handle_errno (errno))
+	  return NULL;
+     }
+   return fp;
 }
+
 static int fclose_fun (FILE *fp)
 {
-   int status = fclose (fp);
-   if (status != 0)
-     _pSLerrno_errno = errno;
-   return status;
+   errno = 0;
+   while (EOF == fclose (fp))
+     {
+	if (0 == handle_errno (errno))
+	  return EOF;
+     }
+   return 0;
 }
 
 static void stdio_fopen (char *file, char *mode)
@@ -365,35 +404,50 @@ static int close_file_type (SL_File_Table_Type *t)
    if (NULL == fp) ret = -1;
    else
      {
-	if (0 == (t->flags & SL_PIPE))
+	while (1)
 	  {
-	     if (EOF == (ret = fclose (fp)))
-	       _pSLerrno_errno = errno;
-	  }
+	     if (0 == (t->flags & SL_PIPE))
+	       {
+		  if (EOF == (ret = fclose (fp)))
+		    ret = -1;
+	       }
 #ifdef HAVE_POPEN
-	else
-	  {
-	     if (-1 == (ret = pclose (fp)))
-	       _pSLerrno_errno = errno;
-	  }
+	     else
+	       {
+		  ret = pclose (fp);
+	       }
 #endif
+	     if ((ret != -1)
+		 || (0 == handle_errno (errno)))
+	       break;
+	  }
      }
-
+   if (t->buf != NULL) SLfree (t->buf);
    if (t->file != NULL) SLang_free_slstring (t->file);
    memset ((char *) t, 0, sizeof (SL_File_Table_Type));
    return ret;
 }
 
-static int stdio_fclose (SL_File_Table_Type *t)
+static int stdio_fclose (void)
 {
    int ret;
+   SLang_MMT_Type *mmt;
+   SL_File_Table_Type *t;
 
+   if (NULL == (mmt = SLang_pop_mmt (SLANG_FILE_PTR_TYPE)))
+     return -1;
+
+   t = (SL_File_Table_Type *) SLang_object_from_mmt (mmt);
    if (NULL == check_fp (t, 0xFFFF))
      return -1;
 
+   if (t->flags & SL_FDOPEN)
+     _pSLfclose_fdopen_fp (mmt);
+
    ret = close_file_type (t);
 
    t->flags = SL_INUSE;
+   SLang_free_mmt (mmt);
    return ret;
 }
 
@@ -643,10 +697,7 @@ static int stdio_fputs (char *s, SL_File_Table_Type *t)
    if (NULL == (fp = check_fp (t, SL_WRITE)))
      return -1;
 
-   if (-1 == signal_safe_fputs (s, fp))
-     return -1;
-
-   return (int) _pSLstring_bytelen (s);
+   return signal_safe_fputs (s, fp);
 }
 
 static int stdio_fflush (SL_File_Table_Type *t)
@@ -655,11 +706,13 @@ static int stdio_fflush (SL_File_Table_Type *t)
 
    if (NULL == (fp = check_fp (t, SL_WRITE)))
      return -1;
-
-   if (EOF == fflush (fp))
+   
+   errno = 0;
+   while ((EOF == fflush (fp)) || ferror (fp))
      {
-	_pSLerrno_errno = errno;
-	return -1;
+	if (0 == handle_errno (errno))
+	  return -1;
+	clearerr (fp);
      }
 
    return 0;
@@ -697,16 +750,21 @@ static void stdio_fread_bytes (SLang_Ref_Type *ref, unsigned int *num_wantedp, S
    int ret = -1;
    char *buf = NULL;
    SLang_BString_Type *bs;
-
    if (NULL == (fp = check_fp (t, SL_READ)))
      goto the_return;
    
    if (NULL == (buf = SLmalloc (num_wanted + 1)))
      goto the_return;
    
-   errno = 0;
-   while (0 == (num_read = fread (buf, sizeof(char), num_wanted, fp)))
+   while (num_read < num_wanted)
      {
+	unsigned int dnum;
+	
+	dnum = fread (buf + num_read, sizeof(char), num_wanted-num_read, fp);
+	num_read += dnum;
+	if (num_read == num_wanted)
+	  break;
+	
 	if (0 == handle_errno (errno))
 	  break;
      }
@@ -714,7 +772,6 @@ static void stdio_fread_bytes (SLang_Ref_Type *ref, unsigned int *num_wantedp, S
    ret = check_ferror_and_realloc (fp, 0, &buf, num_wanted, num_read, sizeof (char));
    if (ret == -1)
      goto the_return;
-   
 
    bs = SLbstring_create_malloced ((unsigned char *)buf, num_read, 1);
    ret = SLang_assign_to_ref (ref, SLANG_BSTRING_TYPE, (VOID_STAR)&bs);
@@ -963,10 +1020,7 @@ static int stdio_fprintf (void)
 	return -1;
      }
    
-   if (-1 == signal_safe_fputs (s, fp))
-     status = -1;
-   else
-     status = (int) _pSLstring_bytelen (s);
+   status = signal_safe_fputs (s, fp);
 
    SLang_free_mmt (mmt);
    _pSLang_free_slstring (s);
@@ -984,16 +1038,66 @@ static int stdio_printf (void)
    if (-1 == SLang_pop_slstring (&s))
      return -1;
    
-   if (-1 == signal_safe_fputs (s, stdout))
-     status = -1;
-   else
-     status = (int) _pSLstring_bytelen (s);
+   status = signal_safe_fputs (s, stdout);
 
    _pSLang_free_slstring (s);
    return status;
 }
 
+#ifdef HAVE_SETVBUF
+static int stdio_setvbuf (SL_File_Table_Type *t, int *modep, int *sizep)
+{
+   FILE *fp;
+   int status;
+   char *newbuf;
+   unsigned int size;
+
+   if (NULL == (fp = check_fp (t, 0xFFFF)))
+     return -1;
    
+   if (*sizep < 0)
+     {
+	SLang_verror (SL_InvalidParm_Error, "setvbuf: Expecting a positive integer for the size parameter");
+	return -1;
+     }
+
+   size = (unsigned int) *sizep;
+   newbuf = NULL;
+
+   errno = 0;
+   if (*modep == _IONBF)
+     status = setvbuf (fp, NULL, _IONBF, 0);
+   else if (*sizep == 0)
+     status = setvbuf (fp, NULL, *modep, 0);
+   else
+     {
+	if (size <= 0) size = BUFSIZ;
+
+	if (NULL == (newbuf = SLmalloc (size)))
+	  return -1;
+	errno = 0;
+	status = setvbuf (fp, newbuf, *modep, size);
+     }
+
+   if (status != 0)
+     {
+	_pSLerrno_errno = errno;
+	if (newbuf != NULL)
+	  SLfree (newbuf);
+	return -1;
+     }
+
+   if (newbuf != NULL)
+     {
+	if (t->buf != NULL)
+	  SLfree (t->buf);
+	t->buf = newbuf;
+	t->buflen = size;
+     }
+   return status;
+}
+#endif
+
 #define F SLANG_FILE_PTR_TYPE
 #define R SLANG_REF_TYPE
 #define I SLANG_INT_TYPE
@@ -1009,7 +1113,7 @@ static SLang_Intrin_Fun_Type Stdio_Name_Table[] =
    MAKE_INTRINSIC_SS("fopen", stdio_fopen, V),
    MAKE_INTRINSIC_1("feof", stdio_feof, I, F),
    MAKE_INTRINSIC_1("ferror", stdio_ferror, I, F),
-   MAKE_INTRINSIC_1("fclose", stdio_fclose, I, F),
+   MAKE_INTRINSIC_0("fclose", stdio_fclose, I),
    MAKE_INTRINSIC_2("fgets", stdio_fgets, I, R, F),
    MAKE_INTRINSIC_1("fflush", stdio_fflush, I, F),
    MAKE_INTRINSIC_2("fputs", stdio_fputs, I, S, F),
@@ -1023,7 +1127,10 @@ static SLang_Intrin_Fun_Type Stdio_Name_Table[] =
    MAKE_INTRINSIC_1("fwrite", stdio_fwrite, V, F),
 #ifdef HAVE_POPEN
    MAKE_INTRINSIC_SS("popen", stdio_popen, V),
-   MAKE_INTRINSIC_1("pclose", stdio_fclose, I, F),
+   MAKE_INTRINSIC_0("pclose", stdio_fclose, I),
+#endif
+#ifdef HAVE_SETVBUF
+   MAKE_INTRINSIC_3("setvbuf", stdio_setvbuf, I, F, I, I),
 #endif
    SLANG_END_INTRIN_FUN_TABLE
 };
@@ -1051,13 +1158,27 @@ static SLang_IConstant_Type Stdio_Consts [] =
    MAKE_ICONSTANT("SEEK_SET", SEEK_SET),
    MAKE_ICONSTANT("SEEK_END", SEEK_END),
    MAKE_ICONSTANT("SEEK_CUR", SEEK_CUR),
+#ifdef _IONBF
+   MAKE_ICONSTANT("_IONBF", _IONBF),
+#endif
+#ifdef _IOLBF
+   MAKE_ICONSTANT("_IOLBF", _IOLBF),
+#endif
+#ifdef _IOFBF
+   MAKE_ICONSTANT("_IOFBF", _IOFBF),
+#endif
    SLANG_END_ICONST_TABLE
 };
 
 static void destroy_file_type (SLtype type, VOID_STAR ptr)
 {
+   /* Do not change errno when a variable goes out of scope */
+   int e = _pSLerrno_errno;
+
    (void) type;
    (void) close_file_type ((SL_File_Table_Type *) ptr);
+
+   _pSLerrno_errno = e;
 }
 
 
